import wsgiref.validate

import pytest
from werkzeug.exceptions import BadRequest

from rolo.request import Request, dummy_wsgi_environment, get_raw_path, restore_payload


def test_get_json():
    r = Request(
        "POST",
        "/",
        headers={"Content-Type": "application/json"},
        body=b'{"foo": "bar", "baz": 420}',
    )
    assert r.json == {"foo": "bar", "baz": 420}
    assert r.content_type == "application/json"


def test_get_json_force():
    r = Request("POST", "/", body=b'{"foo": "bar", "baz": 420}')
    assert r.get_json(force=True) == {"foo": "bar", "baz": 420}


def test_get_json_invalid():
    r = Request("POST", "/", body=b'{"foo": "')

    with pytest.raises(BadRequest):
        assert r.get_json(force=True)

    assert r.get_json(force=True, silent=True) is None


def test_get_data():
    r = Request("GET", "/", body="foobar")
    assert r.data == b"foobar"


def test_get_data_as_text():
    r = Request("GET", "/", body="foobar")
    assert r.get_data(as_text=True) == "foobar"


def test_get_stream():
    r = Request("GET", "/", body=b"foobar")
    assert r.stream.read(3) == b"foo"
    assert r.stream.read(3) == b"bar"


def test_args():
    r = Request("GET", "/", query_string="foo=420&bar=69")
    assert len(r.args) == 2
    assert r.args["foo"] == "420"
    assert r.args["bar"] == "69"


def test_values():
    r = Request("GET", "/", query_string="foo=420&bar=69")
    assert len(r.values) == 2
    assert r.values["foo"] == "420"
    assert r.values["bar"] == "69"


def test_form_empty():
    r = Request("POST", "/")
    assert len(r.form) == 0


def test_post_form_urlencoded_and_query():
    # see https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/POST#example
    r = Request(
        "POST",
        "/form",
        query_string="query1=foo&query2=bar",
        body=b"field1=value1&field2=value2",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
    )

    assert len(r.form) == 2
    assert r.form["field1"] == "value1"
    assert r.form["field2"] == "value2"

    assert len(r.args) == 2
    assert r.args["query1"] == "foo"
    assert r.args["query2"] == "bar"

    assert len(r.values) == 4
    assert r.values["field1"] == "value1"
    assert r.values["field2"] == "value2"
    assert r.args["query1"] == "foo"
    assert r.args["query2"] == "bar"


def test_validate_dummy_environment():
    def validate(*args, **kwargs):
        assert wsgiref.validate.check_environ(dummy_wsgi_environment(*args, **kwargs)) is None

    validate(path="/foo/bar", body="foo")
    validate(path="/foo/bar", query_string="foo=420&bar=69")
    validate(server=("localstack.cloud", 4566))
    validate(server=("localstack.cloud", None))
    validate(remote_addr="127.0.0.1")
    validate(headers={"Content-Type": "text/xml"}, body=b"")
    validate(headers={"Content-Type": "text/xml", "x-amz-target": "foobar"}, body=b"")


def test_content_length_is_set_automatically():
    # checking that the value is calculated automatically
    request = Request("GET", "/", body="foobar")
    assert request.content_length == 6


def test_content_length_is_overwritten():
    # checking that the value passed from headers take precedence
    request = Request("GET", "/", body="foobar", headers={"Content-Length": "7"})
    assert request.content_length == 7


def test_get_custom_headers():
    request = Request("GET", "/", body="foobar", headers={"x-amz-target": "foobar"})
    assert request.headers["x-amz-target"] == "foobar"


def test_get_raw_path():
    request = Request("GET", "/foo/bar/ed", raw_path="/foo%2Fbar/ed")

    assert request.path == "/foo/bar/ed"
    assert request.environ["RAW_URI"] == "/foo%2Fbar/ed"
    assert get_raw_path(request) == "/foo%2Fbar/ed"


def test_get_raw_path_with_query():
    request = Request("GET", "/foo/bar/ed", raw_path="/foo%2Fbar/ed?fizz=buzz")

    assert request.path == "/foo/bar/ed"
    assert request.environ["RAW_URI"] == "/foo%2Fbar/ed?fizz=buzz"
    assert get_raw_path(request) == "/foo%2Fbar/ed"


def test_get_raw_path_with_prefix_slashes():
    request = Request("GET", "/foo/bar/ed", raw_path="//foo%2Fbar/ed?fizz=buzz")

    assert request.path == "/foo/bar/ed"
    assert request.environ["RAW_URI"] == "//foo%2Fbar/ed?fizz=buzz"
    assert get_raw_path(request) == "//foo%2Fbar/ed"


def test_get_raw_path_with_full_uri():
    # raw_path is actually raw_uri in the WSGI environment
    # it can be a full URL
    request = Request("GET", "/foo/bar/ed", raw_path="http://localhost:4566/foo%2Fbar/ed")

    assert request.path == "/foo/bar/ed"
    assert request.environ["RAW_URI"] == "http://localhost:4566/foo%2Fbar/ed"
    assert get_raw_path(request) == "/foo%2Fbar/ed"


def test_headers_retain_dashes():
    request = Request("GET", "/foo/bar/ed", {"X-Amz-Meta--foo_bar-ed": "foobar"})
    assert "x-amz-meta--foo_bar-ed" in request.headers
    assert request.headers["x-amz-meta--foo_bar-ed"] == "foobar"


def test_headers_retain_case():
    request = Request("GET", "/foo/bar/ed", {"X-Amz-Meta--FOO_BaR-ed": "foobar"})
    keys = list(request.headers.keys())
    for k in keys:
        if k.lower().startswith("x-amz-meta"):
            assert k == "X-Amz-Meta--FOO_BaR-ed"
            return
    pytest.fail(f"key not in header keys {keys}")


def test_multipart_parsing():
    body = (
        b"--4efd159eae0c4f4e125a5a509e073d85"
        b"\r\n"
        b'Content-Disposition: form-data; name="foo"; filename="foo"'
        b"\r\n\r\n"
        b"bar"
        b"\r\n"
        b"--4efd159eae0c4f4e125a5a509e073d85"
        b"\r\n"
        b'Content-Disposition: form-data; name="baz"; filename="baz"'
        b"\r\n\r\n"
        b"ed"
        b"\r\n--4efd159eae0c4f4e125a5a509e073d85--"
        b"\r\n"
    )

    request = Request(
        "POST",
        path="/",
        body=body,
        headers={"Content-Type": "multipart/form-data; boundary=4efd159eae0c4f4e125a5a509e073d85"},
    )
    result = {}
    for k, file_storage in request.files.items():
        result[k] = file_storage.stream.read().decode("utf-8")

    assert result == {"foo": "bar", "baz": "ed"}


def test_utf8_path():
    r = Request("GET", "/foo/Ā0Ä")

    assert r.path == "/foo/Ā0Ä"
    assert r.environ["PATH_INFO"] == "/foo/Ä\x800Ã\x84"  # quoted and latin-1 encoded


def test_restore_payload_multipart_parsing():
    body = (
        b"\r\n"
        b"--4efd159eae0c4f4e125a5a509e073d85"
        b"\r\n"
        b'Content-Disposition: form-data; name="formfield"'
        b"\r\n\r\n"
        b"not a file, just a field"
        b"\r\n"
        b"--4efd159eae0c4f4e125a5a509e073d85"
        b"\r\n"
        b'Content-Disposition: form-data; name="foo"; filename="foo"'
        b"\r\n"
        b"Content-Type: text/plain;"
        b"\r\n\r\n"
        b"bar"
        b"\r\n"
        b"--4efd159eae0c4f4e125a5a509e073d85"
        b"\r\n"
        b'Content-Disposition: form-data; name="baz"; filename="baz"'
        b"\r\n"
        b"Content-Type: text/plain;"
        b"\r\n\r\n"
        b"ed"
        b"\r\n"
        b"\r\n--4efd159eae0c4f4e125a5a509e073d85--"
        b"\r\n"
    )

    request = Request(
        "POST",
        path="/",
        body=body,
        headers={"Content-Type": "multipart/form-data; boundary=4efd159eae0c4f4e125a5a509e073d85"},
    )

    form = {}
    for k, field in request.form.items():
        form[k] = field

    assert form == {"formfield": "not a file, just a field"}

    files = []
    for k, file_storage in request.files.items():
        assert file_storage.stream
        # we do not want to consume the file storage stream, because we can't restore the payload then
        files.append(k)

    assert files == ["foo", "baz"]
    restored_data = restore_payload(request)

    assert restored_data == body


def test_request_mixed_multipart():
    # this is almost how we previously restored a form that would have both `form` fields and `files`
    # we would URL encode the form first then add multipart, which does not work, the first part should be ignored
    # and make certain strict multipart parser fail (Starlette), because it finds data before the first boundary

    # this test does something a bit different to prove it is ignored (add an URL encoded part in the beginning)
    body = (
        b"formfield=not+a+file%2C+just+a+field\r\n"
        b"--4efd159eae0c4f4e125a5a509e073d85"
        b"\r\n"
        b'Content-Disposition: form-data; name="foo"; filename="foo"'
        b"\r\n"
        b"Content-Type: text/plain;"
        b"\r\n\r\n"
        b"bar"
        b"\r\n"
        b"--4efd159eae0c4f4e125a5a509e073d85"
        b"\r\n"
        b'Content-Disposition: form-data; name="baz"; filename="baz"'
        b"\r\n"
        b"Content-Type: text/plain;"
        b"\r\n\r\n"
        b"ed"
        b"\r\n"
        b"\r\n--4efd159eae0c4f4e125a5a509e073d85--"
        b"\r\n"
    )

    request = Request(
        "POST",
        path="/",
        body=body,
        headers={"Content-Type": "multipart/form-data; boundary=4efd159eae0c4f4e125a5a509e073d85"},
    )

    form = {}
    for k, field in request.form.items():
        form[k] = field

    assert form == {}

    files = []
    for k, file_storage in request.files.items():
        assert file_storage.stream
        # we do not want to consume the file storage stream, because we can't restore the payload then
        files.append(k)

    assert files == ["foo", "baz"]

    restored_data = restore_payload(request)
    assert b"formfield" not in restored_data


def test_restore_payload_form_urlencoded():
    body = b"formfield=not+a+file%2C+just+a+field"

    request = Request(
        "POST",
        path="/",
        body=body,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
    )

    form = {}
    for k, field in request.form.items():
        form[k] = field

    assert form == {"formfield": "not a file, just a field"}

    assert not request.files

    restored_data = restore_payload(request)

    assert restored_data == body



"""Several gateway handlers"""
import typing as t

from werkzeug.datastructures import Headers
from werkzeug.exceptions import HTTPException, NotFound

from rolo.response import Response
from rolo.routing import Router

from .chain import HandlerChain, RequestContext


class RouterHandler:
    """
    Adapter to serve a ``Router`` as a ``Handler``.
    """

    router: Router
    respond_not_found: bool

    def __init__(self, router: Router, respond_not_found: bool = False) -> None:
        self.router = router
        self.respond_not_found = respond_not_found

    def __call__(self, chain: HandlerChain, context: RequestContext, response: Response):
        try:
            router_response = self.router.dispatch(context.request)
            response.update_from(router_response)
            chain.stop()
        except NotFound:
            if self.respond_not_found:
                chain.respond(404, "not found")


class EmptyResponseHandler:
    """
    Handler that creates a default response if the response in the context is empty.
    """

    status_code: int
    body: bytes
    headers: dict

    def __init__(self, status_code: int = 404, body: bytes = None, headers: Headers = None):
        self.status_code = status_code
        self.body = body or b""
        self.headers = headers or Headers()

    def __call__(self, chain: HandlerChain, context: RequestContext, response: Response):
        if self.is_empty_response(response):
            self.populate_default_response(response)

    def is_empty_response(self, response: Response):
        return response.status_code in [0, None] and not response.response

    def populate_default_response(self, response: Response):
        response.status_code = self.status_code
        response.data = self.body
        response.headers.update(self.headers)


class WerkzeugExceptionHandler:
    def __init__(self, output_format: t.Literal["json", "html"] = None) -> None:
        self.format = output_format or "json"

    def __call__(
        self, chain: HandlerChain, exception: Exception, context: RequestContext, response: Response
    ):
        if not isinstance(exception, HTTPException):
            return

        headers = Headers(exception.get_headers())  # FIXME
        headers.pop()

        if self.format == "html":
            chain.respond(status_code=exception.code, headers=headers, payload=exception.get_body())
        elif self.format == "json":
            chain.respond(
                status_code=exception.code,
                headers=headers,
                payload={"code": exception.code, "description": exception.description},
            )
        else:
            raise ValueError(f"unknown rendering format {self.format}")



from .converter import PortConverter, RegexConverter
from .handler import handler_dispatcher
from .router import RequestArguments, Router, route
from .rules import RuleAdapter, RuleGroup, WithHost

__all__ = [
    "PortConverter",
    "RegexConverter",
    "RequestArguments",
    "Router",
    "RuleAdapter",
    "RuleGroup",
    "WithHost",
    "handler_dispatcher",
    "route",
]



import abc
from urllib.parse import urlparse

import requests
from werkzeug import Request, Response
from werkzeug.datastructures import Headers

from .request import get_raw_base_url, get_raw_current_url, get_raw_path, restore_payload


class HttpClient(abc.ABC):
    """
    An HTTP client that can make http requests using werkzeug's request object.
    """

    @abc.abstractmethod
    def request(self, request: Request, server: str | None = None) -> Response:
        """
        Make the given HTTP as a client.

        :param request: the request to make
        :param server: the URL to send the request to, which defaults to the host component of the original Request.
        :return: the response.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def close(self):
        """
        Close any underlying resources the client may need.
        """
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class _VerifyRespectingSession(requests.Session):
    """
    A class which wraps requests.Session to circumvent https://github.com/psf/requests/issues/3829.
    This ensures that if `REQUESTS_CA_BUNDLE` or `CURL_CA_BUNDLE` are set, the request does not perform the TLS
    verification if `session.verify` is set to `False.
    """

    def merge_environment_settings(self, url, proxies, stream, verify, *args, **kwargs):
        if self.verify is False:
            verify = False

        return super(_VerifyRespectingSession, self).merge_environment_settings(
            url, proxies, stream, verify, *args, **kwargs
        )


class SimpleRequestsClient(HttpClient):
    session: requests.Session
    follow_redirects: bool

    def __init__(self, session: requests.Session = None, follow_redirects: bool = True):
        self.session = session or _VerifyRespectingSession()
        self.follow_redirects = follow_redirects

    @staticmethod
    def _get_destination_url(request: Request, server: str | None = None) -> str:
        if server:
            # accepts "http://localhost:5000" or "localhost:5000"
            if "://" in server:
                parts = urlparse(server)
                scheme, server = parts.scheme, parts.netloc
            else:
                scheme = request.scheme
            return get_raw_current_url(scheme, server, request.root_path, get_raw_path(request))

        return get_raw_base_url(request)

    def request(self, request: Request, server: str | None = None) -> Response:
        """
        Very naive implementation to make the given HTTP request using the requests library, i.e., process the request
        as a client.

        :param request: the request to perform
        :param server: the URL to send the request to, which defaults to the host component of the original Request.
        :param allow_redirects: allow the request to follow redirects
        :return: the response.
        """

        url = self._get_destination_url(request, server)

        headers = dict(request.headers.items())

        # urllib3 (used by requests) will set an Accept-Encoding header ("gzip,deflate")
        # - See urllib3.util.request.ACCEPT_ENCODING
        # - This could be circumvented by setting Accept-Encoding explicitly to None
        # BUT: Python's http client (used by urllib3) will always add an accept header if none is given
        # - See https://github.com/psf/requests/issues/2234 and http.client.putrequest
        # Explicitly set `Accept-Encoding: identity` here if no `Accept-Encoding` header is set in the originating
        # request to avoid any unused manipulations by underlying libraries.
        if not request.headers.get("accept-encoding"):
            headers["accept-encoding"] = "identity"

        response = self.session.request(
            method=request.method,
            # use raw base url to preserve path url encoding
            url=url,
            # request.args are only the url parameters
            params=list(request.args.items(multi=True)),
            headers=headers,
            data=restore_payload(request),
            stream=True,
            allow_redirects=self.follow_redirects,
        )

        if request.method == "HEAD":
            # for HEAD  requests we have to keep the original content-length, but it will be re-calculated when creating
            # the final_response object
            final_response = Response(
                response=response.content,
                status=response.status_code,
                headers=Headers(dict(response.headers)),
            )
            final_response.content_length = response.headers.get("Content-Length", 0)
            return final_response

        response_headers = Headers(dict(response.headers))
        if "chunked" in (transfer_encoding := response_headers.get("Transfer-Encoding", "")):
            response_headers.pop("Content-Length", None)
            # We should not set `Transfer-Encoding` in a Response, because it is the responsibility of the webserver
            # to do so, if there are no Content-Length. However, gzip behavior is more related to the actual content of
            # the response, so we keep that one.
            transfer_encoding_values = [v.strip() for v in transfer_encoding.split(",")]
            transfer_encoding_no_chunked = [
                v for v in transfer_encoding_values if v.lower() != "chunked"
            ]
            response_headers.setlist("Transfer-Encoding", transfer_encoding_no_chunked)

        final_response = Response(
            response=(chunk for chunk in response.raw.stream(1024, decode_content=False)),
            status=response.status_code,
            headers=response_headers,
        )

        return final_response

    def close(self):
        self.session.close()


def make_request(request: Request) -> Response:
    """
    Convenience method to make the given HTTP as a client.

    :param request: the request to make
    :return: the response.
    """
    with SimpleRequestsClient() as client:
        return client.request(request)



from .chain import (
    CompositeExceptionHandler,
    CompositeFinalizer,
    CompositeHandler,
    CompositeResponseHandler,
    ExceptionHandler,
    Handler,
    HandlerChain,
    RequestContext,
)
from .gateway import Gateway

__all__ = [
    "Gateway",
    "HandlerChain",
    "RequestContext",
    "Handler",
    "ExceptionHandler",
    "CompositeHandler",
    "CompositeExceptionHandler",
    "CompositeResponseHandler",
    "CompositeFinalizer",
]



import logging
import typing as t

from ..request import Request
from ..response import Response
from ..websocket.request import WebSocketRequest
from .chain import ExceptionHandler, Handler, HandlerChain, RequestContext

LOG = logging.getLogger(__name__)


class Gateway:
    """
    A gateway creates new HandlerChain instances for each request and processes requests through them.
    """

    request_handlers: list[Handler]
    response_handlers: list[Handler]
    finalizers: list[Handler]
    exception_handlers: list[ExceptionHandler]

    def __init__(
        self,
        request_handlers: list[Handler] = None,
        response_handlers: list[Handler] = None,
        finalizers: list[Handler] = None,
        exception_handlers: list[ExceptionHandler] = None,
        context_class: t.Type[RequestContext] = None,
    ) -> None:
        super().__init__()
        self.request_handlers = request_handlers if request_handlers is not None else []
        self.response_handlers = response_handlers if response_handlers is not None else []
        self.exception_handlers = exception_handlers if exception_handlers is not None else []
        self.finalizers = finalizers if finalizers is not None else []
        self.context_class = context_class or RequestContext

    def new_chain(self) -> HandlerChain:
        return HandlerChain(
            self.request_handlers,
            self.response_handlers,
            self.finalizers,
            self.exception_handlers,
        )

    def process(self, request: Request, response: Response):
        chain = self.new_chain()

        context = self.context_class(request)

        chain.handle(context, response)

    def accept(self, request: WebSocketRequest):
        response = Response(status=101)
        self.process(request, response)

        # only send the populated response if the websocket hasn't already done so before
        if response.status_code != 101:
            if request.is_upgraded():
                return
            if request.is_rejected():
                return
            request.reject(response)



import inspect
import typing as t

from werkzeug.routing import Map, Rule, RuleFactory


class _RuleAttributes(t.NamedTuple):
    path: str
    host: t.Optional[str] = (None,)
    methods: t.Optional[t.Iterable[str]] = None
    kwargs: t.Optional[dict[str, t.Any]] = {}


class _RouteEndpoint(t.Protocol):
    """
    An endpoint that encapsulates ``_RuleAttributes`` for the creation of a ``Rule`` inside a ``Router``.
    """

    rule_attributes: list[_RuleAttributes]

    def __call__(self, *args, **kwargs):
        raise NotImplementedError


class WithHost(RuleFactory):
    def __init__(self, host: str, rules: t.Iterable[RuleFactory]) -> None:
        self.host = host
        self.rules = rules

    def get_rules(self, map: Map) -> t.Iterator[Rule]:
        for rulefactory in self.rules:
            for rule in rulefactory.get_rules(map):
                rule = rule.empty()
                rule.host = self.host
                yield rule


class RuleGroup(RuleFactory):
    def __init__(self, rules: t.Iterable[RuleFactory]):
        self.rules = rules

    def get_rules(self, map: Map) -> t.Iterable[Rule]:
        for rule in self.rules:
            yield from rule.get_rules(map)


class RuleAdapter(RuleFactory):
    """
    Takes something that can also be passed to ``Router.add``, and exposes it as a ``RuleFactory`` that generates the
    appropriate Werkzeug rules. This can be used in combination with other rule factories like ``Submount``,
    and creates general compatibility with werkzeug rules. Here's an example::

        @route("/my_api", methods=["GET"])
        def do_get(request: Request, _args):
            # should be inherited
            return Response(f"{request.path}/do-get")

        def hello(request: Request, _args):
            return Response(f"hello world")

        router = Router()

        # base endpoints
        endpoints = RuleAdapter([
            do_get,
            RuleAdapter("/hello", hello)
        ])

        router.add([
            endpoints,
            Submount("/foo", [endpoints])
        ])

    """

    factory: RuleFactory
    """The underlying real rule factory."""

    @t.overload
    def __init__(
        self,
        path: str,
        endpoint: t.Callable,
        host: t.Optional[str] = None,
        methods: t.Optional[t.Iterable[str]] = None,
        **kwargs,
    ):
        """
        Basically a ``Rule``.

        :param path: the path pattern to match. This path rule, in contrast to the default behavior of Werkzeug, will be
                        matched against the raw / original (potentially URL-encoded) path.
        :param endpoint: the endpoint to invoke
        :param host: an optional host matching pattern. if not pattern is given, the rule matches any host
        :param methods: the allowed HTTP verbs for this rule
        :param kwargs: any other argument that can be passed to ``werkzeug.routing.Rule``
        """
        ...

    @t.overload
    def __init__(self, fn: _RouteEndpoint):
        """
        Takes a route endpoint (typically a function decorated with ``@route``) and adds it as ``EndpointRule``.

        :param fn: the RouteEndpoint function
        """
        ...

    @t.overload
    def __init__(self, rule_factory: RuleFactory):
        """
        Adds a ``Rule`` or the rules created by a ``RuleFactory`` to the given router. It passes the rules down to
        the underlying Werkzeug ``Map``, but also returns the created Rules.

        :param rule_factory: a `Rule` or ``RuleFactory`
        """
        ...

    @t.overload
    def __init__(self, obj: t.Any):
        """
        Scans the given object for members that can be used as a `RouteEndpoint` and adds them to the router.

        :param obj: the object to scan
        """
        ...

    @t.overload
    def __init__(self, rules: list[t.Union[_RouteEndpoint, RuleFactory, t.Any]]):
        """Add multiple rules at once"""
        ...

    def __init__(self, *args, **kwargs):
        """
        Dispatcher for overloaded ``__init__`` methods.
        """
        if "path" in kwargs or isinstance(args[0], str):
            self.factory = _EndpointRule(*args, **kwargs)
        elif "fn" in kwargs or callable(args[0]):
            self.factory = _EndpointFunction(*args, **kwargs)
        elif "rule_factory" in kwargs:
            self.factory = kwargs["rule_factory"]
        elif isinstance(args[0], RuleFactory):
            self.factory = args[0]
        elif isinstance(args[0], list):
            self.factory = RuleGroup([RuleAdapter(rule) for rule in args[0]])
        else:
            self.factory = _EndpointsObject(*args, **kwargs)

    def get_rules(self, map: Map) -> t.Iterable[Rule]:
        yield from self.factory.get_rules(map)


class _EndpointRule(RuleFactory):
    """
    Generates default werkzeug ``Rule`` object with the given attributes. Additionally, it makes sure that
    the generated rule always has a default host value, if the map has host matching enabled. Specifically,
    it adds the well-known placeholder ``<__host__>``, which is later stripped out of the request arguments
    when dispatching to the endpoint. This ensures compatibility of rule definitions across routers that
    have host matching enabled or not.
    """

    def __init__(
        self,
        path: str,
        endpoint: t.Callable,
        host: t.Optional[str] = None,
        methods: t.Optional[t.Iterable[str]] = None,
        **kwargs,
    ):
        self.path = path
        self.endpoint = endpoint
        self.host = host
        self.methods = methods
        self.kwargs = kwargs

    def get_rules(self, map: Map) -> t.Iterable[Rule]:
        host = self.host

        if host is None and map.host_matching:
            # this creates a "match any" rule, and will put the value of the host
            # into the variable "__host__"
            host = "<__host__>"

        # the typing for endpoint is a str, but the doc states it can be any value,
        # however then the redirection URL building will not work
        rule = Rule(
            self.path, endpoint=self.endpoint, methods=self.methods, host=host, **self.kwargs
        )
        yield rule


class _EndpointFunction(RuleFactory):
    """
    Internal rule factory that generates router Rules from ``@route`` annotated functions, or anything else
    that can be interpreted as a ``_RouteEndpoint``. It extracts the rule attributes from the
    ``_RuleAttributes`` attribute defined by ``_RouteEndpoint``. Example::

        @route("/my_api", methods=["GET"])
        def do_get(request: Request, _args):
            # should be inherited
            return Response(f"{request.path}/do-get")

        router.add(do_get)  # <- will use an _EndpointFunction RuleFactory.
    """

    def __init__(self, fn: _RouteEndpoint):
        self.fn = fn

    def get_rules(self, map: Map) -> t.Iterable[Rule]:
        attrs: list[_RuleAttributes] = self.fn.rule_attributes
        for attr in attrs:
            yield from _EndpointRule(
                path=attr.path,
                endpoint=self.fn,
                host=attr.host,
                methods=attr.methods,
                **attr.kwargs,
            ).get_rules(map)


class _EndpointsObject(RuleFactory):
    """
    Scans the given object for members that can be used as a `RouteEndpoint` and yields them as rules.
    """

    def __init__(self, obj: object):
        self.obj = obj

    def get_rules(self, map: Map) -> t.Iterable[Rule]:
        endpoints: list[_RouteEndpoint] = []

        members = inspect.getmembers(self.obj)
        for _, member in members:
            if hasattr(member, "rule_attributes"):
                endpoints.append(member)

        # make sure rules with "HEAD" are added first, otherwise werkzeug would let any "GET" rule would overwrite them.
        for endpoint in endpoints:
            for attr in endpoint.rule_attributes:
                if attr.methods and "HEAD" in attr.methods:
                    yield from _EndpointRule(
                        path=attr.path,
                        endpoint=endpoint,
                        host=attr.host,
                        methods=attr.methods,
                        **attr.kwargs,
                    ).get_rules(map)

        for endpoint in endpoints:
            for attr in endpoint.rule_attributes:
                if not attr.methods or "HEAD" not in attr.methods:
                    yield from _EndpointRule(
                        path=attr.path,
                        endpoint=endpoint,
                        host=attr.host,
                        methods=attr.methods,
                        **attr.kwargs,
                    ).get_rules(map)



import gzip
import json

import pytest
import requests
from pytest_httpserver import HTTPServer
from werkzeug import Request as WerkzeugRequest
from werkzeug.datastructures import Headers

from rolo import Request, Response
from rolo.client import SimpleRequestsClient
from rolo.proxy import Proxy, ProxyHandler, forward


@pytest.fixture
def router_server(wsgi_router_server):
    """Creates a new Router with a handler dispatcher, serves it through a newly created server, and returns
    both the router and the server.
    """
    yield wsgi_router_server


def echo_request_metadata_handler(request: WerkzeugRequest) -> Response:
    """
    Simple request handler that returns the incoming request metadata (method, path, url, headers).

    :param request: the incoming HTTP request
    :return: an HTTP response
    """
    response = Response()
    response.set_json(
        {
            "method": request.method,
            "path": request.path,
            "url": request.url,
            "headers": dict(Headers(request.headers)),
        }
    )
    return response


class TestPathForwarder:
    def test_get_with_path_rule(self, router_server, httpserver: HTTPServer):
        router, proxy = router_server
        backend = httpserver

        backend.expect_request("/").respond_with_data("ok/")
        backend.expect_request("/bar").respond_with_data("ok/bar")
        backend.expect_request("/bar/ed").respond_with_data("ok/bar/ed")

        router.add("/foo/<path:path>", ProxyHandler(backend.url_for("/")))

        response = requests.get(proxy.url + "/foo/bar")
        assert response.ok
        assert response.text == "ok/bar"

        response = requests.get(proxy.url + "/foo/bar/ed")
        assert response.ok
        assert response.text == "ok/bar/ed"

        response = requests.get(proxy.url)
        assert not response.ok

        response = requests.get(proxy.url + "/bar")
        assert not response.ok

        backend.check()

    def test_get_with_plain_rule(self, router_server, httpserver: HTTPServer):
        router, proxy = router_server
        backend = httpserver

        backend.expect_request("/").respond_with_data("ok")

        router.add("/foo", ProxyHandler(backend.url_for("/")))

        response = requests.get(proxy.url + "/foo")
        assert response.ok
        assert response.text == "ok"

        response = requests.get(proxy.url + "/foo/bar")
        assert not response.ok

    def test_get_with_different_base_url(self, router_server, httpserver: HTTPServer):
        router, proxy = router_server
        backend = httpserver

        backend.expect_request("/bar/ed").respond_with_data("ok/bar/ed")
        backend.expect_request("/bar/ed/baz").respond_with_data("ok/bar/ed/baz")

        router.add("/foo/<path:path>", ProxyHandler(backend.url_for("/bar")))

        response = requests.get(proxy.url + "/foo/ed")
        assert response.ok
        assert response.text == "ok/bar/ed"

        response = requests.get(proxy.url + "/foo/ed/baz")
        assert response.ok
        assert response.text == "ok/bar/ed/baz"

    def test_get_with_different_base_url_plain_rule(self, router_server, httpserver: HTTPServer):
        router, proxy = router_server
        backend = httpserver

        backend.expect_request("/bar").respond_with_data("ok/bar")
        backend.expect_request("/bar/").respond_with_data("ok/bar/")

        router.add("/foo", ProxyHandler(backend.url_for("/bar")))

        response = requests.get(proxy.url + "/foo")
        assert response.ok
        assert response.text == "ok/bar/"  # it's calling /bar/ because it's part of the root URL

    def test_xff_header(self, router_server, httpserver: HTTPServer):
        router, proxy = router_server
        backend = httpserver

        def _echo_headers(request):
            return Response(json.dumps(dict(request.headers)), mimetype="application/json")

        backend.expect_request("/echo").respond_with_handler(_echo_headers)

        router.add("/<path:path>", ProxyHandler(backend.url_for("/")))

        response = requests.get(proxy.url + "/echo")
        assert response.ok
        headers = response.json()
        assert headers["X-Forwarded-For"] == "127.0.0.1"

        # check that it appends remote address correctly if a header is already present
        response = requests.get(proxy.url + "/echo", headers={"X-Forwarded-For": "127.0.0.2"})
        assert response.ok
        headers = response.json()
        assert headers["X-Forwarded-For"] == "127.0.0.2, 127.0.0.1"

    def test_post_form_data_with_query_args(self, router_server, httpserver: HTTPServer):
        router, proxy = router_server
        backend = httpserver

        def _handler(request: WerkzeugRequest):
            data = {
                "args": request.args,
                "form": request.form,
            }
            return Response(json.dumps(data), mimetype="application/json")

        backend.expect_request("/form").respond_with_handler(_handler)

        router.add("/<path:path>", ProxyHandler(backend.url_for("/")))

        response = requests.post(
            proxy.url + "/form?q=yes",
            data={"foo": "bar", "baz": "ed"},
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )
        assert response.ok
        doc = response.json()
        assert doc == {"args": {"q": "yes"}, "form": {"foo": "bar", "baz": "ed"}}

    def test_path_encoding_preservation(self, router_server, httpserver: HTTPServer):
        router, proxy = router_server
        backend = httpserver

        def _handler(request: WerkzeugRequest):
            from rolo.request import get_raw_path

            data = {"path": get_raw_path(request), "query": request.query_string.decode("utf-8")}
            return Response(json.dumps(data), mimetype="application/json")

        backend.expect_request("").respond_with_handler(_handler)

        router.add("/<path:path>", ProxyHandler(backend.url_for("/")))

        response = requests.get(
            proxy.url
            + "/arn%3Aaws%3Aservice%3Aeu-west-1%3A000000000000%3Aroot-arn-path%2Fsub-arn-path%2F%2A/%E4%B8%8A%E6%B5%B7%2B%E4%B8%AD%E5%9C%8B?%E4%B8%8A%E6%B5%B7%2B%E4%B8%AD%E5%9C%8B=%E4%B8%8A%E6%B5%B7%2B%E4%B8%AD%E5%9C%8B",
        )
        assert response.ok
        doc = response.json()
        assert doc == {
            "path": "/arn%3Aaws%3Aservice%3Aeu-west-1%3A000000000000%3Aroot-arn-path%2Fsub-arn-path%2F%2A/%E4%B8%8A%E6%B5%B7%2B%E4%B8%AD%E5%9C%8B",
            "query": "%E4%B8%8A%E6%B5%B7%2B%E4%B8%AD%E5%9C%8B=%E4%B8%8A%E6%B5%B7%2B%E4%B8%AD%E5%9C%8B",
        }

    @pytest.mark.parametrize("chunked", [True, False])
    def test_proxy_handler_transfer_encoding(self, router_server, httpserver: HTTPServer, chunked):
        router, proxy = router_server
        backend = httpserver
        body = "enough-for-content-length"

        def _handler(_: WerkzeugRequest):
            # if the response is chunked, return a generator instead, which will return `Transfer-Encoding: chunked`
            if chunked:
                _body = (c for c in body)
            else:
                _body = body

            return Response(_body, status=200)

        backend.expect_request("").respond_with_handler(_handler)

        router.add("/", ProxyHandler(backend.url_for("/")))

        response = requests.get(proxy.url)

        if chunked:
            assert response.headers["Transfer-Encoding"] == "chunked"
            assert "Content-Length" not in response.headers
        else:
            assert response.headers["Content-Length"] == str(len(body))
            assert "Transfer-Encoding" not in response.headers

        assert response.text == body


class TestProxy:
    def test_proxy_with_custom_client(self, httpserver: HTTPServer):
        """The Proxy class allows the injection of a custom HTTP client which can attach default headers to every
        request. this test verifies that this works through the proxy implementation."""
        httpserver.expect_request("/").respond_with_handler(echo_request_metadata_handler)

        with SimpleRequestsClient() as client:
            client.session.headers["X-My-Custom-Header"] = "hello world"

            proxy = Proxy(httpserver.url_for("/").lstrip("/"), client)

            request = Request(
                path="/",
                method="POST",
                body="foobar",
                remote_addr="127.0.0.10",
                headers={"Host": "127.0.0.1:80"},
            )

            response = proxy.request(request)

            assert "X-My-Custom-Header" in response.json["headers"]
            assert response.json["method"] == "POST"
            assert response.json["headers"]["X-My-Custom-Header"] == "hello world"
            assert response.json["headers"]["X-Forwarded-For"] == "127.0.0.10"
            assert response.json["headers"]["Host"] == "127.0.0.1:80"

    @pytest.mark.parametrize("chunked", [True, False])
    def test_proxy_for_transfer_encoding_chunked(
        self,
        httpserver: HTTPServer,
        chunked,
    ):
        body = "enough-for-content-length"

        def _handler(_request: Request) -> Response:
            # if the response is chunked, return a generator instead, which will return `Transfer-Encoding: chunked`
            if chunked:
                _body = (c for c in body)
            else:
                _body = body

            return Response(_body, status=200)

        httpserver.expect_request("").respond_with_handler(_handler)

        proxy = Proxy(httpserver.url_for("/").lstrip("/"))

        request = Request(path="/", method="GET", headers={"Host": "127.0.0.1:80"})

        response = proxy.request(request)

        if chunked:
            # the proxy should not return a Transfer-Encoding, as this is something the webserver should set
            assert "Transfer-Encoding" not in response.headers
            assert "Content-Length" not in response.headers
        else:
            assert response.headers["Content-Length"] == str(len(body))
            assert "Transfer-Encoding" not in response.headers

        assert response.data.decode() == body

    @pytest.mark.parametrize(
        "chunked,gzipped",
        [
            (False, False),
            (False, True),
            (True, False),
            (True, True),
        ],
    )
    def test_proxy_for_transfer_encoding_chunked_and_gzip(
        self,
        httpserver: HTTPServer,
        chunked,
        gzipped,
    ):
        body = b"enough-for-content-length"
        if gzipped:
            body = gzip.compress(body, mtime=0)

        def _handler(_request: Request) -> Response:
            # if the response is chunked, return a generator instead, which will return `Transfer-Encoding: chunked`
            headers = {}
            _body = body
            if gzipped:
                headers["Transfer-Encoding"] = "gzip"

            if chunked:
                _body = (chr(c).encode("latin-1") for c in body)

            return Response(_body, status=200, headers=headers)

        httpserver.expect_request("/proxy").respond_with_handler(_handler)

        proxy = Proxy(httpserver.url_for("/").lstrip("/"))

        request = Request(path="/proxy", method="GET", headers={"Host": "127.0.0.1:80"})

        response = proxy.request(request)

        if gzipped:
            assert response.headers["Transfer-Encoding"] == "gzip"

        if chunked:
            assert "Content-Length" not in response.headers
            assert "chunked" not in response.headers.get("Transfer-Encoding", "")
        else:
            assert response.headers["Content-Length"] == str(len(body))
            if not gzipped:
                assert "Transfer-Encoding" not in response.headers

        assert response.data == body


@pytest.mark.parametrize("consume_data", [True, False])
def test_forward_files_and_form_data_proxy_consumes_data(
    consume_data, serve_asgi_adapter, tmp_path
):
    """Tests that, when the proxy consumes (or doesn't consume) the request object's data prior to forwarding,
    the request is forwarded correctly. not using httpserver here because it consumes werkzeug data incorrectly (it
    calls ``request.get_data()``)."""

    @WerkzeugRequest.application
    def _backend_handler(request: WerkzeugRequest):
        data = {
            "data": request.data.decode("utf-8"),
            "args": request.args,
            "form": request.form,
            "files": {
                name: storage.stream.read().decode("utf-8")
                for name, storage in request.files.items()
            },
        }
        return Response(json.dumps(data), mimetype="application/json")

    @WerkzeugRequest.application
    def _proxy_handler(request: WerkzeugRequest):
        # heuristic to check whether the stream has been consumed
        assert getattr(request, "_cached_data", None) is None, "data has already been cached"

        if consume_data:
            assert (
                not request.data
            )  # data should be empty because it is consumed by parsing form data

        return forward(request, forward_base_url=backend.url)

    backend = serve_asgi_adapter(_backend_handler)
    proxy = serve_asgi_adapter(_proxy_handler)

    tmp_file_1 = tmp_path / "temp_file_1.txt"
    tmp_file_1.write_text("1: hello\nworld")

    tmp_file_2 = tmp_path / "temp_file_2.txt"
    tmp_file_2.write_text("2: foobar")

    response = requests.post(
        proxy.url,
        params={"q": "yes"},
        data={"foo": "bar", "baz": "ed"},
        files={"upload_file_1": open(tmp_file_1, "rb"), "upload_file_2": open(tmp_file_2, "rb")},
    )
    assert response.ok
    doc = response.json()
    assert doc == {
        "data": "",
        "args": {"q": "yes"},
        "form": {"foo": "bar", "baz": "ed"},
        "files": {
            "upload_file_1": "1: hello\nworld",
            "upload_file_2": "2: foobar",
        },
    }



from .adapter import WebSocketEnvironment, WebSocketListener
from .errors import WebSocketDisconnectedError, WebSocketError, WebSocketProtocolError
from .request import WebSocket, WebSocketRequest

__all__ = [
    "WebSocket",
    "WebSocketDisconnectedError",
    "WebSocketEnvironment",
    "WebSocketError",
    "WebSocketListener",
    "WebSocketProtocolError",
    "WebSocketRequest",
]



import functools
import threading
import typing as t
from typing import overload

from werkzeug import Request, Response
from werkzeug.routing import BaseConverter, Map, Rule, RuleFactory

from rolo.request import get_raw_path

from .converter import PortConverter, RegexConverter
from .rules import RuleAdapter, _RouteEndpoint, _RuleAttributes

if t.TYPE_CHECKING:
    from _typeshed.wsgi import WSGIApplication

HTTP_METHODS = ("GET", "POST", "PUT", "PATCH", "DELETE", "HEAD", "OPTIONS", "TRACE")

E = t.TypeVar("E")
RequestArguments = t.Mapping[str, t.Any]


class Dispatcher(t.Protocol[E]):
    """
    A Dispatcher is called when a URL route matches a request. The dispatcher is responsible for appropriately
    creating a Response from the incoming Request and the matching endpoint.
    """

    def __call__(self, request: Request, endpoint: E, args: RequestArguments) -> Response:
        """
        Dispatch the HTTP Request.

        :param request: the incoming HTTP request
        :param endpoint: the endpoint that matched the URL rule
        :param args: the request arguments extracted from the URL rule
        :return: an HTTP Response
        """
        pass


def route(
    path: str, host: t.Optional[str] = None, methods: t.Optional[t.Iterable[str]] = None, **kwargs
) -> t.Callable[[E], list[_RouteEndpoint]]:
    """
    Decorator that indicates that the given function is a Router Rule.

    :param path: the path pattern to match
    :param host: an optional host matching pattern. if not pattern is given, the rule matches any host
    :param methods: the allowed HTTP methods for this rule
    :param kwargs: any other argument that can be passed to ``werkzeug.routing.Rule``
    :return: the function endpoint wrapped as a ``_RouteEndpoint``
    """

    def wrapper(fn: E):
        if hasattr(fn, "rule_attributes"):
            route_marker = fn
        else:

            @functools.wraps(fn)
            def route_marker(*args, **kwargs):
                return fn(*args, **kwargs)

            route_marker.rule_attributes = []

        route_marker.rule_attributes.append(_RuleAttributes(path, host, methods, kwargs))

        return route_marker

    return wrapper


def call_endpoint(
    request: Request,
    endpoint: t.Callable[[Request, RequestArguments], Response],
    args: RequestArguments,
) -> Response:
    """
    A Dispatcher that treats the matching endpoint as a callable and invokes it with the Request and request arguments.
    """
    return endpoint(request, args)


def _clone_map_without_rules(old: Map) -> Map:
    return Map(
        default_subdomain=old.default_subdomain,
        strict_slashes=old.strict_slashes,
        merge_slashes=old.merge_slashes,
        redirect_defaults=old.redirect_defaults,
        converters=old.converters,
        sort_parameters=old.sort_parameters,
        sort_key=old.sort_key,
        host_matching=old.host_matching,
    )


def _clone_map_with_rules(old: Map) -> Map:
    """
    Creates a new copy of the existing map, with fresh unbound copies of all its containing rules.

    :param old: the map to copy
    :return: a new instance of the map
    """
    new = _clone_map_without_rules(old)

    for old_rule in old.iter_rules():
        new.add(old_rule.empty())

    return new


class Router(t.Generic[E]):
    """
    A Router is a wrapper around werkzeug's routing Map, that adds convenience methods and additional dispatching
    logic via the ``Dispatcher`` Protocol.
    """

    default_converters: dict[str, t.Type[BaseConverter]] = {
        "regex": RegexConverter,
        "port": PortConverter,
    }

    url_map: Map
    dispatcher: Dispatcher[E]

    def __init__(
        self,
        dispatcher: Dispatcher[E] = None,
        converters: t.Mapping[str, t.Type[BaseConverter]] = None,
    ):
        if converters is None:
            converters = dict(self.default_converters)
        else:
            converters = {**self.default_converters, **converters}

        self.url_map = Map(
            host_matching=True,
            strict_slashes=False,
            converters=converters,
            redirect_defaults=False,
        )
        self.dispatcher = dispatcher or call_endpoint
        self._mutex = threading.RLock()

    @overload
    def add(
        self,
        path: str,
        endpoint: E,
        host: t.Optional[str] = None,
        methods: t.Optional[t.Iterable[str]] = None,
        **kwargs,
    ) -> Rule:
        """
        Creates a new Rule from the given parameters and adds it to the URL Map.

        TODO: many callers still expect ``add`` to return a single rule rather than a list, but it would be better to
         homogenize the API and make every method return a list.

        :param path: the path pattern to match. This path rule, in contrast to the default behavior of Werkzeug, will be
                        matched against the raw / original (potentially URL-encoded) path.
        :param endpoint: the endpoint to invoke
        :param host: an optional host matching pattern. if not pattern is given, the rule matches any host
        :param methods: the allowed HTTP verbs for this rule
        :param kwargs: any other argument that can be passed to ``werkzeug.routing.Rule``
        :return: the rule that was created
        """
        ...

    @overload
    def add(self, fn: _RouteEndpoint) -> list[Rule]:
        """
        Adds a RouteEndpoint (typically a function decorated with ``@route``) as a rule to the router.

        :param fn: the RouteEndpoint function
        :return: the rules that were added. for this operation, only one rule will be in the list
        """
        ...

    @overload
    def add(self, rule_factory: RuleFactory) -> list[Rule]:
        """
        Adds a ``Rule`` or the rules created by a ``RuleFactory`` to the given router. It passes the rules down to
        the underlying Werkzeug ``Map``, but also returns the created Rules.

        :param rule_factory: a `Rule` or ``RuleFactory`
        :return: the rules that were added
        """
        ...

    @overload
    def add(self, obj: t.Any) -> list[Rule]:
        """
        Scans the given object for members that can be used as a `RouteEndpoint` and adds them to the router.

        :param obj: the object to scan
        :return: the rules that were added
        """
        ...

    @overload
    def add(self, routes: list[t.Union[_RouteEndpoint, RuleFactory, t.Any]]) -> list[Rule]:
        """
        Add multiple routes or rules to the router at once.

        :param routes: the objects to add as routes
        :return: the rules that were added
        """
        ...

    def add(self, *args, **kwargs) -> t.Union[Rule, list[Rule]]:
        """
        Creates a ``RuleAdapter`` and adds the generated rules to the router's Map.
        """
        rules = self._add_rules(RuleAdapter(*args, **kwargs))

        if "path" in kwargs or isinstance(args[0], str):
            return rules[0]

        return rules

    def _add_rules(self, rule_factory: RuleFactory) -> list[Rule]:
        """
        Thread safe version of Werkzeug's ``Map.add``.

        :param rule_factory: the rule to add
        """
        with self._mutex:
            new = _clone_map_with_rules(self.url_map)

            # instantiate, modify, and collect rules
            rules = []
            for rule in rule_factory.get_rules(new):
                rules.append(rule)

                if rule.host is None and new.host_matching:
                    # this creates a "match any" rule, and will put the value of the host
                    # into the variable "__host__"
                    rule.host = "<__host__>"

            for rule in rules:
                new.add(rule)

            self.url_map = new

            return rules

    def add_rule(self, rule: RuleFactory):
        """
        Thread safe version of Werkzeug's ``Map.add``. This can be used as low-level method to pass a rule directly
        to the Werkzeug URL map without any manipulation or manual creation of the rule, which ``add`` does. Like
        ``remove``, the method actually clones and replaces the underlying URL Map, to guarantee thread safety with
        ``dispatch``. Adding rules is therefore a relatively expensive operation.

        :param rule: the rule to add
        """
        with self._mutex:
            new = _clone_map_with_rules(self.url_map)
            new.add(rule)
            self.url_map = new

    def remove(self, rules: t.Union[Rule, t.Iterable[Rule]]):
        """
        Removes a single Rule from the Router.

        **Caveat**: This is an expensive operation. Removing rules from a URL Map is intentionally not supported by
        werkzeug due to issues with thread safety, see https://github.com/pallets/werkzeug/issues/796, and because
        using a lock in ``match`` would be too expensive. However, some services that use Routers for routing
        internal resources need to be able to remove rules when those resources are removed. So to remove rules we
        create a new Map without that rule. This will not prevent the rules from dispatching until the Map has been
        completely constructed.

        :param rules: the Rule or rules to remove that were previously returned by ``add``.
        """
        if isinstance(rules, Rule):
            self._remove_rules([rules])
        else:
            self._remove_rules(rules)

    def _remove_rules(self, rules: t.Iterable[Rule]):
        """
        Removes a set of Rules from the Router.

        **Caveat**: This is an expensive operation. Removing rules from a URL Map is intentionally not supported by
        werkzeug due to issues with thread safety, see https://github.com/pallets/werkzeug/issues/796, and because
        using a lock in ``match`` would be too expensive. However, some services that use Routers for routing
        internal resources need to be able to remove rules when those resources are removed. So to remove rules we
        create a new Map without that rule. This will not prevent the rules from dispatching until the Map has been
        completely constructed.

        :param rules: the list of Rule objects to remove that were previously returned by ``add``.
        """
        with self._mutex:
            old = self.url_map
            for r in rules:
                if r not in old._rules:
                    raise KeyError("no such rule")

            # collect all old rules that are not in the set of rules to remove
            new = _clone_map_without_rules(old)

            for old_rule in old.iter_rules():
                if old_rule in rules:
                    # this works even with copied rules because of the __eq__ implementation of Rule
                    continue

                new.add(old_rule.empty())
            self.url_map = new

    def dispatch(self, request: Request) -> Response:
        """
        Does the entire dispatching roundtrip, from matching the request to endpoints, and then invoking the endpoint
        using the configured dispatcher of the router. For more information on the matching behavior,
        see ``werkzeug.routing.MapAdapter.match()``.

        :param request: the HTTP request
        :return: the HTTP response
        """
        matcher = self.url_map.bind(server_name=request.host)
        # Match on the _raw_ path to ensure that converters (like "path") can extract the raw path.
        # f.e. router.add(/<path:path>, ProxyHandler(...))
        # If we would use the - already url-decoded - request.path here, a handler would not be able to access
        # the original (potentially URL-encoded) path.
        # As a consequence, rules need to match on URL-encoded URLs (f.e. use '%20' instead of ' ').
        handler, args = matcher.match(get_raw_path(request), method=request.method)
        args.pop("__host__", None)
        return self.dispatcher(request, handler, args)

    def route(
        self,
        path: str,
        host: t.Optional[str] = None,
        methods: t.Optional[t.Iterable[str]] = None,
        **kwargs,
    ) -> t.Callable[[E], _RouteEndpoint]:
        """
        Returns a ``route`` decorator and immediately adds it to the router instance. This effectively mimics flask's
        ``@app.route``.

        :param path: the path pattern to match
        :param host: an optional host matching pattern. if not pattern is given, the rule matches any host
        :param methods: the allowed HTTP verbs for this rule
        :param kwargs: any other argument that can be passed to ``werkzeug.routing.Rule``
        :return: the function endpoint wrapped as a ``_RouteEndpoint``
        """

        def wrapper(fn):
            r = route(path, host, methods, **kwargs)
            fn = r(fn)
            self.add(fn)
            return fn

        return wrapper

    def wsgi(self) -> "WSGIApplication":
        """
        Returns this router as a WSGI compatible interface. This can be used to conveniently serve a Router instance
        through a WSGI server, for instance werkzeug's dev server::

            from werkzeug.serving import run_simple

            from rolo import Router
            from rolo.dispatcher import handler_dispatcher

            router = Router(dispatcher=handler_dispatcher())
            run_simple("localhost", 5000, router.wsgi())

        :return: a WSGI callable that invokes this router
        """
        return Request.application(self.dispatch)



"""
DEPRECATED: use ``rolo.routing`` instead.
"""

from .routing.converter import PortConverter, RegexConverter
from .routing.router import RequestArguments, Router, route
from .routing.rules import RuleAdapter, RuleGroup, WithHost

__all__ = [
    "Router",
    "route",
    "RegexConverter",
    "PortConverter",
    "RuleAdapter",
    "RuleGroup",
    "WithHost",
    "RequestArguments",
]



from typing import TypedDict

import pydantic
import pytest
from werkzeug.exceptions import BadRequest

from rolo import Request, Router, resource
from rolo.routing import handler as routing_handler
from rolo.routing import handler_dispatcher

pydantic_version = pydantic.version.version_short()


class MyItem(pydantic.BaseModel):
    name: str
    price: float
    is_offer: bool = None


class TestPydanticHandlerDispatcher:
    def test_request_arg(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler(_request: Request, item: MyItem) -> dict:
            return {"item": item.model_dump()}

        router.add("/items", handler)

        request = Request("POST", "/items", body=b'{"name":"rolo","price":420.69}')
        assert router.dispatch(request).get_json(force=True) == {
            "item": {
                "name": "rolo",
                "price": 420.69,
                "is_offer": None,
            },
        }

    def test_request_args(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler(_request: Request, item_id: int, item: MyItem) -> dict:
            return {"item_id": item_id, "item": item.model_dump()}

        router.add("/items/<int:item_id>", handler)

        request = Request("POST", "/items/123", body=b'{"name":"rolo","price":420.69}')
        assert router.dispatch(request).get_json(force=True) == {
            "item_id": 123,
            "item": {
                "name": "rolo",
                "price": 420.69,
                "is_offer": None,
            },
        }

    def test_request_args_empty_body(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler(_request: Request, item_id: int, item: MyItem) -> dict:
            return {"item_id": item_id, "item": item.model_dump()}

        router.add("/items/<int:item_id>", handler)

        request = Request("POST", "/items/123", body=b"")
        assert router.dispatch(request).get_json(force=True) == [
            {
                "type": "json_invalid",
                "loc": [],
                "msg": "Invalid JSON: EOF while parsing a value at line 1 column 0",
                "ctx": {"error": "EOF while parsing a value at line 1 column 0"},
                "input": "",
                "url": f"https://errors.pydantic.dev/{pydantic_version}/v/json_invalid",
            }
        ]

    def test_response(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler(_request: Request, item_id: int) -> MyItem:
            return MyItem(name="rolo", price=420.69)

        router.add("/items/<int:item_id>", handler)

        request = Request("GET", "/items/123")
        assert router.dispatch(request).get_json() == {
            "name": "rolo",
            "price": 420.69,
            "is_offer": None,
        }

    def test_response_list(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler(_request: Request) -> list[MyItem]:
            return [
                MyItem(name="rolo", price=420.69),
                MyItem(name="twiks", price=1.23, is_offer=True),
            ]

        router.add("/items", handler)

        request = Request("GET", "/items")
        assert router.dispatch(request).get_json() == [
            {
                "name": "rolo",
                "price": 420.69,
                "is_offer": None,
            },
            {
                "name": "twiks",
                "price": 1.23,
                "is_offer": True,
            },
        ]

    def test_request_arg_validation_error(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler(_request: Request, item_id: int, item: MyItem) -> str:
            return item.model_dump_json()

        router.add("/items/<int:item_id>", handler)

        request = Request("POST", "/items/123", body=b'{"name":"rolo"}')
        assert router.dispatch(request).get_json() == [
            {
                "type": "missing",
                "loc": ["price"],
                "msg": "Field required",
                "input": {"name": "rolo"},
                "url": f"https://errors.pydantic.dev/{pydantic_version}/v/missing",
            }
        ]

    def test_request_arg_invalid_json(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler(_request: Request, item_id: int, item: MyItem) -> str:
            return item.model_dump_json()

        router.add("/items/<int:item_id>", handler)

        request = Request("POST", "/items/123", body=b'{"}')
        with pytest.raises(BadRequest):
            assert router.dispatch(request)

    def test_missing_annotation(self):
        router = Router(dispatcher=handler_dispatcher())

        # without an annotation, we cannot be sure what type to serialize into, so the dispatcher doesn't pass
        # anything into ``item``.
        def handler(_request: Request, item=None) -> dict:
            return {"item": item}

        router.add("/items", handler)

        request = Request("POST", "/items", body=b'{"name":"rolo","price":420.69}')
        assert router.dispatch(request).get_json(force=True) == {"item": None}

    def test_with_pydantic_disabled(self, monkeypatch):
        monkeypatch.setattr(routing_handler, "ENABLE_PYDANTIC", False)
        router = Router(dispatcher=handler_dispatcher())

        def handler(_request: Request, item: MyItem) -> dict:
            return {"item": item.model_dump()}

        router.add("/items", handler)

        request = Request("POST", "/items", body=b'{"name":"rolo","price":420.69}')
        with pytest.raises(TypeError):
            # "missing 1 required positional argument: 'item'"
            assert router.dispatch(request)

    def test_with_resource(self):
        router = Router(dispatcher=handler_dispatcher())

        @resource("/items/<int:item_id>")
        class MyResource:
            def on_get(self, request: Request, item_id: int):
                return MyItem(name="rolo", price=420.69)

            def on_post(self, request: Request, item_id: int, item: MyItem):
                return {"item_id": item_id, "item": item.model_dump()}

        router.add(MyResource())

        response = router.dispatch(Request("GET", "/items/123"))
        assert response.get_json() == {
            "name": "rolo",
            "price": 420.69,
            "is_offer": None,
        }

        response = router.dispatch(
            Request("POST", "/items/123", body=b'{"name":"rolo","price":420.69}')
        )
        assert response.get_json() == {
            "item": {"is_offer": None, "name": "rolo", "price": 420.69},
            "item_id": 123,
        }

    def test_with_generic_type_alias(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler(request: Request, matrix: dict[str, str] = None):
            return "ok"

        router.add("/", endpoint=handler)

        request = Request("GET", "/")
        assert router.dispatch(request).data == b"ok"

    def test_with_typed_dict(self):
        try:
            from typing import Unpack
        except ImportError:
            pytest.skip("This test only works with Python >=3.11")

        router = Router(dispatcher=handler_dispatcher())

        class Test(TypedDict, total=False):
            path: str
            random_value: str

        def func(request: Request, **kwargs: Unpack[Test]):
            return f"path={kwargs.get('path')},random_value={kwargs.get('random_value')}"

        router.add(
            "/",
            endpoint=func,
            defaults={"path": "", "random_value": "dev"},
        )

        request = Request("GET", "/")
        assert router.dispatch(request).data == b"path=,random_value=dev"



import inspect
import typing as t

import pydantic

from rolo.request import Request
from rolo.response import Response

from .handler import Handler, HandlerDispatcher, ResultValue
from .router import RequestArguments


def _get_model_argument(endpoint: Handler) -> t.Optional[tuple[str, t.Type[pydantic.BaseModel]]]:
    """
    Inspects the endpoint function using Python reflection to find in its signature a ``pydantic.BaseModel`` attribute.

    :param endpoint: the endpoint to inspect
    :return: a tuple containing the name and class, or None
    """
    if not inspect.isfunction(endpoint) and not inspect.ismethod(endpoint):
        # cannot yet dispatch to other callables (e.g. an object with a `__call__` method)
        return None

    # finds the first pydantic.BaseModel in the list of annotations.
    # ``def foo(request: Request, id: int, item: MyItem)`` would yield ``('my_item', MyItem)``
    for arg_name, arg_type in endpoint.__annotations__.items():
        if arg_name in ("self", "return"):
            continue
        if not inspect.isclass(arg_type):
            continue
        try:
            if issubclass(arg_type, pydantic.BaseModel):
                return arg_name, arg_type
        except TypeError:
            # FIXME: this is needed for Python 3.10 support
            continue

    return None


def _try_parse_pydantic_request_body(
    request: Request, endpoint: Handler
) -> t.Optional[tuple[str, pydantic.BaseModel]]:
    arg = _get_model_argument(endpoint)

    if not arg:
        return

    arg_name, arg_type = arg

    if not request.content_length:
        # forces a ValidationError "Invalid JSON: EOF while parsing a value at line 1 column 0"
        arg_type.model_validate_json(b"")

    # will raise a werkzeug.BadRequest error if the JSON is invalid
    obj = request.get_json(force=True)

    return arg_name, arg_type.model_validate(obj)


class PydanticHandlerDispatcher(HandlerDispatcher):
    """
    Special HandlerDispatcher that knows how to serialize and deserialize pydantic models.
    """

    def invoke_endpoint(
        self,
        request: Request,
        endpoint: t.Callable,
        request_args: RequestArguments,
    ) -> t.Any:
        # prepare request args
        try:
            arg = _try_parse_pydantic_request_body(request, endpoint)
        except pydantic.ValidationError as e:
            return Response(e.json(), mimetype="application/json", status=400)

        if arg:
            arg_name, model = arg
            request_args = {**request_args, arg_name: model}

        return super().invoke_endpoint(request, endpoint, request_args)

    def populate_response(self, response: Response, value: ResultValue):
        # try to convert any pydantic types to dicts before handing them to the parent implementation
        if isinstance(value, pydantic.BaseModel):
            value = value.model_dump()
        elif isinstance(value, (list, tuple)):
            converted = []
            for element in value:
                if isinstance(element, pydantic.BaseModel):
                    converted.append(element.model_dump())
                else:
                    converted.append(element)
            value = converted

        super().populate_response(response, value)



import typing as t

from werkzeug.routing import BaseConverter, Map


class RegexConverter(BaseConverter):
    """
    A converter that can be used to inject a regex as parameter, e.g., ``path=/<regex('[a-z]+'):my_var>``.
    When using groups in regex, make sure they are non-capturing ``(?:[a-z]+)``
    """

    def __init__(self, map: Map, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(map, *args, **kwargs)
        self.regex = args[0]


class PortConverter(BaseConverter):
    """
    Useful to optionally match ports for host patterns, like ``localstack.localhost.cloud<port:port>``. Notice how you
    don't need to specify the colon. The regex matches it if the port is there, and will remove the colon if matched.
    The converter converts the port to an int, or returns None if there's no port in the input string.
    """

    regex = r"(?::[0-9]{1,5})?"

    def to_python(self, value: str) -> t.Any:
        if value:
            return int(value[1:])
        return None



import functools
import typing as t

from werkzeug import Response
from werkzeug._internal import _wsgi_decoding_dance
from werkzeug.datastructures import EnvironHeaders, Headers, MultiDict
from werkzeug.sansio.request import Request as _SansIORequest
from werkzeug.wsgi import _get_server

from .adapter import (
    BytesMessage,
    CreateConnection,
    Message,
    TextMessage,
    WebSocketAdapter,
    WebSocketEnvironment,
    WebSocketListener,
)
from .errors import WebSocketDisconnectedError, WebSocketProtocolError


class WebSocket:
    """
    High-level interface to interact with a websocket after a handshake has been completed with
    `WebsocketRequest.accept()`.
    """

    request: "WebSocketRequest"
    socket: WebSocketAdapter

    def __init__(self, request: "WebSocketRequest", socket: WebSocketAdapter):
        self.request = request
        self.socket = socket

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __iter__(self):
        while True:
            try:
                yield self.receive()
            except WebSocketDisconnectedError:
                break

    def send(self, text_or_bytes: str | bytes, timeout: float = None):
        """
        Send data to the websocket connection.

        :param text_or_bytes: the data to send. Use strings for text-mode sockets (default).
        :param timeout: the timeout in seconds to wait before raising a timeout error
        """
        if text_or_bytes is None:
            raise ValueError("text_or_bytes cannot be None")

        if isinstance(text_or_bytes, str):
            event = TextMessage(text_or_bytes)
        elif isinstance(text_or_bytes, bytes):
            event = BytesMessage(text_or_bytes)
        else:
            raise WebSocketProtocolError(
                f"Cannot send data type {type(text_or_bytes)} over websocket"
            )
        self.socket.send(event, timeout)

    def receive(self) -> str | bytes:
        """
        Receive the next data package from the websocket. Will be string or byte data and set the
        underlying binary for the frame automatically.

        :raise WebSocketDisconnectedError: if the websocket was closed in the meantime
        :raise WebSocketProtocolError: error in the interaction between the app and the webserver
        :return: the next data package from the websocket
        """
        event = self.socket.receive()
        if isinstance(event, Message):
            data = event.data
            if data is None:
                raise WebSocketProtocolError("No data returned by the websocket.")
            return data
        else:
            raise WebSocketProtocolError(
                f"Unexpected websocket event type {event.__class__.__name__}."
            )

    def close(self, code: int = 1000, reason: t.Optional[str] = None, timeout: float = None):
        """
        Closes the websocket connection with specific code.

        :param code: the websocket close code.
        :param reason: optional reason
        :param timeout: connection timeout
        """
        self.socket.close(code, reason, timeout)


class WebSocketRequest(_SansIORequest):
    """
    A websocket request represents the incoming HTTP request to upgrade the connection to a WebSocket
    connection. The request method is an artificial ``WEBSOCKET`` method that can also be used in the Router:
    ``@route("/path", method=["WEBSOCKET"])``.

    The websocket connection needs to be either accepted or rejected. When calling
    ``WebSocketRequest.accept``, an upgrade response will be sent to the client, and the protocol will be
    switched to the bidirectional WebSocket protocol. If ``WebSocketRequest.reject`` is called, the server
    immediately returns an HTTP response and closes the connection.
    """

    def __init__(self, environ: WebSocketEnvironment):
        """
        Creates a new request from the given WebSocketEnvironment. This is like a sans-IO WSGI Environment,
        with an additional field ``rolo.websocket`` that contains a ``WebSocketAdapter`` interface.

        :param environ: the WebSocketEnvironment
        """
        raw_headers = environ.get("rolo.headers")
        if raw_headers:
            # restores raw headers from the server scope, to have proper casing or dashes. This can depend on server
            # behavior, but we want a unified way to keep header casing/formatting.
            # This is similar to what we do in wsgi.py
            headers = Headers(
                MultiDict([(k.decode("latin-1"), v.decode("latin-1")) for (k, v) in raw_headers])
            )
        else:
            headers = Headers(EnvironHeaders(environ))

        # copied from werkzeug.wrappers.request
        super().__init__(
            method=environ.get("REQUEST_METHOD", "WEBSOCKET"),
            scheme=environ.get("wsgi.url_scheme", "ws"),
            server=_get_server(environ),
            root_path=_wsgi_decoding_dance(environ.get("SCRIPT_NAME") or ""),
            path=_wsgi_decoding_dance(environ.get("PATH_INFO") or ""),
            query_string=environ.get("QUERY_STRING", "").encode("latin1"),
            headers=headers,
            remote_addr=environ.get("REMOTE_ADDR"),
        )
        self.environ = environ

        self.shallow = True  # compatibility with werkzeug.Request

        self._upgraded = False
        self._rejected = False

    @property
    def socket(self) -> WebSocketAdapter:
        """
        Returns the underlying WebSocketAdapter from the environment. This is analogous to ``Request.stream``
        in the default werkzeug HTTP request object.

        :return: the WebSocketAdapter from the environment
        """
        return self.environ.get("rolo.websocket") or self.environ.get("asgi.websocket")

    def is_upgraded(self) -> bool:
        """Returns true if ``accept`` was called."""
        return self._upgraded

    def is_rejected(self) -> bool:
        """Returns true if ``reject`` was called."""
        return self._rejected

    def reject(self, response: Response):
        """
        Reject the websocket upgrade and return the given response. Will raise a ``ValueError`` if the
        request has already been accepted or rejected before.

        :param response: the HTTP response to return to the client.
        """
        if self._upgraded:
            raise ValueError("Websocket connection already upgraded")
        if self._rejected:
            raise ValueError("Websocket connection already rejected")

        self.socket.reject(
            response.status_code,
            response.headers,
            response.iter_encoded(),
        )
        self._rejected = True

    def accept(
        self, subprotocol: str = None, headers: Headers = None, timeout: float = None
    ) -> WebSocket:
        """
        Performs the websocket connection upgrade handshake. After calling ``accept``, a new ``Websocket``
        instance is returned that represents the bidirectional communication channel, which you should
        continue operating on. Example::

            def app(request: WebsocketRequest):
                # example: do authorization first
                auth = request.headers.get("Authorization")
                if not is_authorized(auth):
                    request.reject(Response("no dice", 403))
                    return

                # then continue working with the websocket
                with request.accept() as websocket:
                    websocket.send("hello world!")
                    data = websocket.receive()
                    # ...

        The handshake using the WebSocketAdapter works as follows: receive the ``CreateConnection`` event
        from the websocket and then call the ``socket.accept(...)``. If the handshake failed because
        the websocket sent an unexpected exception, the connection is closed and the method raises an error.

        :param subprotocol: The subprotocol the server wishes to accept. Optional
        :param headers: Response headers
        :param timeout: connection timeout
        :return: a websocket
        :raises ProtocolError: if unexpected events were received from the websocket server
        """
        if self._upgraded:
            raise ValueError("Websocket connection already upgraded")
        if self._rejected:
            raise ValueError("Websocket connection already rejected")

        event = self.socket.receive(timeout)
        if isinstance(event, CreateConnection):
            self.socket.accept(subprotocol, [], headers, timeout)
            self._upgraded = True
            return WebSocket(self, self.socket)
        else:
            reason = f"Unexpected event {event.__class__.__name__}"
            self.socket.close(1003, reason)
            raise WebSocketProtocolError(reason)

    def close(self):
        """
        Explicitly close the websocket. If this is called after ``reject(...)`` or ``accept(...)`` has been
        called, this will have no effect. Calling ``reject`` inherently closes the websocket connection
        since it immediately returns an HTTP response. After calling ``accept`` you should call
        ``WebSocket.close`` instead.
        """
        if self._rejected or self._upgraded:
            return
        self.socket.close(1000)

    @classmethod
    def listener(cls, fn: t.Callable[["WebSocketRequest"], None]) -> WebSocketListener:
        """
        Convenience function inspired by ``werkzeug.Request.application`` that transforms a function into a
        ``WebsocketListener`` for the use in server code that support ``WebsocketListeners``. Example::

            @WebsocketRequest.listener
            def app(request: WebSocketRequest):
                with request.accept() as ws:
                    ws.send("hello world")

            adapter = ASGIAdapter(wsgi_app=..., websocket_listener=app)
            # ... serve adapter


        :param fn: the function to wrap
        :return: a WebsocketListener compatible interface
        """
        from werkzeug.exceptions import HTTPException

        @functools.wraps(fn)
        def application(*args):
            request = cls(args[-1])
            try:
                fn(*args[:-1] + (request,))
            except HTTPException as e:
                resp = e.get_response(args[-1])
                request.reject(resp)
            finally:
                request.close()

        return application



from .request import Request
from .response import Response
from .routing.resource import Resource, resource
from .routing.router import Router, route

__all__ = [
    "route",
    "resource",
    "Resource",
    "Router",
    "Response",
    "Request",
]



import json
import logging
import typing as t

from werkzeug import Response as WerkzeugResponse

try:
    import pydantic  # noqa

    ENABLE_PYDANTIC = True
except ImportError:
    ENABLE_PYDANTIC = False

from rolo.request import Request
from rolo.response import Response

from .router import Dispatcher, RequestArguments

LOG = logging.getLogger(__name__)

ResultValue = t.Union[
    WerkzeugResponse,
    str,
    bytes,
    dict[str, t.Any],  # a JSON dict
    list[t.Any],
]


class Handler(t.Protocol):
    """
    A protocol used by a ``Router`` together with the dispatcher created with ``handler_dispatcher``. Endpoints added
    with this protocol take as first argument the HTTP request object, and then as keyword arguments the request
    parameters added in the rule. This makes it work very similar to flask routes.

    Example code could look like this::

        def my_route(request: Request, organization: str, repo: str):
            return {"something": "returned as json response"}

        router = Router(dispatcher=handler_dispatcher)
        router.add("/<organization>/<repo>", endpoint=my_route)

    """

    def __call__(self, request: Request, **kwargs) -> ResultValue:
        """
        Handle the given request.

        :param request: the HTTP request object
        :param kwargs: the url request parameters
        :return: a string or bytes value, a dict to create a json response, or a raw werkzeug Response object.
        """
        raise NotImplementedError


class HandlerDispatcher:
    def __init__(self, json_encoder: t.Type[json.JSONEncoder] = None):
        self.json_encoder = json_encoder

    def __call__(
        self, request: Request, endpoint: t.Callable, request_args: RequestArguments
    ) -> Response:
        result = self.invoke_endpoint(request, endpoint, request_args)
        return self.to_response(result)

    def invoke_endpoint(
        self,
        request: Request,
        endpoint: t.Callable,
        request_args: RequestArguments,
    ) -> t.Any:
        return endpoint(request, **request_args)

    def to_response(self, value: ResultValue) -> Response:
        if isinstance(value, WerkzeugResponse):
            return value

        response = Response()
        if value is None:
            return response

        self.populate_response(response, value)
        return response

    def populate_response(self, response: Response, value: ResultValue):
        if isinstance(value, (str, bytes, bytearray)):
            response.data = value
        elif isinstance(value, (dict, list)):
            response.data = json.dumps(value, cls=self.json_encoder)
            response.mimetype = "application/json"
        else:
            raise ValueError("unhandled result type %s", type(value))


def handler_dispatcher(json_encoder: t.Type[json.JSONEncoder] = None) -> Dispatcher[Handler]:
    """
    Creates a Dispatcher that treats endpoints like callables of the ``Handler`` Protocol.

    :param json_encoder: optionally the json encoder class to use for translating responses
    :return: a new dispatcher
    """
    if ENABLE_PYDANTIC:
        from rolo.routing.pydantic import PydanticHandlerDispatcher

        return PydanticHandlerDispatcher(json_encoder)

    return HandlerDispatcher(json_encoder)



"""
The core concepts of the HandlerChain.
"""
from __future__ import annotations

import logging
import typing as t

from werkzeug.datastructures import Headers

from rolo.request import Request
from rolo.response import Response

LOG = logging.getLogger(__name__)


class RequestContext:
    """
    A request context holds the original incoming HTTP Request and arbitrary data. It is passed through the handler
    chain and allows handlers to communicate.
    """

    request: Request

    def __init__(self, request: Request = None):
        self.request = request

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __getattr__(self, item):
        try:
            return self.__dict__[item]
        except KeyError:
            pass
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{item}")

    def get(self, key: str) -> t.Optional[t.Any]:
        return self.__dict__.get(key)


RC = t.TypeVar("RC", bound=RequestContext)


class Handler(t.Protocol[RC]):
    """The signature of request or response handler in the handler chain. Receives the HandlerChain, the
    RequestContext, and the Response object to be populated."""

    def __call__(self, chain: "HandlerChain", context: RC, response: Response):
        ...


class ExceptionHandler(t.Protocol[RC]):
    """The signature of request or response handler in the handler chain. Receives the HandlerChain, the
    RequestContext, and the Response object to be populated."""

    def __call__(
        self, chain: "HandlerChain", exception: Exception, context: RC, response: Response
    ):
        ...


def call_safe(
    func: t.Callable, args: tuple = None, kwargs: dict = None, exception_message: str = None
) -> t.Optional[t.Any]:
    """
    Call the given function with the given arguments, and if it fails, log the given exception_message.
    If logging.DEBUG is set for the logger, then we also log the traceback.

    :param func: function to call
    :param args: arguments to pass
    :param kwargs: keyword arguments to pass
    :param exception_message: message to log on exception
    :return: whatever the func returns
    """
    if exception_message is None:
        exception_message = "error calling function %s" % func.__name__
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}

    try:
        return func(*args, **kwargs)
    except Exception as e:
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.exception(exception_message)
        else:
            LOG.warning("%s: %s", exception_message, e)


class HandlerChain(t.Generic[RC]):
    """
    Implements a variant of the chain-of-responsibility pattern to process an incoming HTTP request. A handler
    chain consists of request handlers, response handlers, finalizers, and exception handlers. Each request
    should have its own HandlerChain instance, since the handler chain holds state for the handling of a
    request. A chain can be in three states that can be controlled by the handlers.

    * Running - the implicit state where all handlers are executed sequentially
    * Stopped - a handler has called ``chain.stop()``. This stops the execution of all request handlers, and
      proceeds immediately to executing the response handlers. Response handlers and finalizers will be run,
      even if the chain has been stopped.
    * Terminated - a handler has called ``chain.terminate()`. This stops the execution of all request
      handlers, and all response handlers, but runs the finalizers at the end.

    If an exception occurs during the execution of request handlers, the chain by default stops the chain,
    then runs each exception handler, and finally runs the response handlers. Exceptions that happen during
    the execution of response or exception handlers are logged but do not modify the control flow of the
    chain.
    """

    # handlers
    request_handlers: list[Handler]
    response_handlers: list[Handler]
    finalizers: list[Handler]
    exception_handlers: list[ExceptionHandler]

    # behavior configuration
    stop_on_error: bool = True
    """If set to true, the chain will implicitly stop if an error occurs in a request handler."""
    raise_on_error: bool = False
    """If set to true, an exception in the request handler will be re-raised by ``handle`` after the exception
    handlers have been called. """

    # internal state
    stopped: bool
    terminated: bool
    error: t.Optional[Exception]
    response: t.Optional[Response]
    context: t.Optional[RequestContext]

    def __init__(
        self,
        request_handlers: list[Handler] = None,
        response_handlers: list[Handler] = None,
        finalizers: list[Handler] = None,
        exception_handlers: list[ExceptionHandler] = None,
    ) -> None:
        super().__init__()
        self.request_handlers = request_handlers if request_handlers is not None else []
        self.response_handlers = response_handlers if response_handlers is not None else []
        self.exception_handlers = exception_handlers if exception_handlers is not None else []
        self.finalizers = finalizers if finalizers is not None else []

        self.stopped = False
        self.terminated = False
        self.finalized = False
        self.error = None
        self.response = None
        self.context = None

    def handle(self, context: RC, response: Response):
        """
        Process the given request and populate the given response according to the handler chain control flow
        described in the ``HandlerChain`` class doc.

        :param context: the incoming request
        :param response: the response to be populated
        """
        self.context = context
        self.response = response

        try:
            for handler in self.request_handlers:
                try:
                    handler(self, self.context, response)
                except Exception as e:
                    # prepare the continuation behavior, but exception handlers could overwrite it
                    if self.raise_on_error:
                        self.error = e
                    if self.stop_on_error:
                        self.stopped = True

                    # call exception handlers safely
                    self._call_exception_handlers(e, response)

                # decide next step
                if self.error:
                    raise self.error
                if self.terminated:
                    return
                if self.stopped:
                    break

            # call response filters
            self._call_response_handlers(response)
        finally:
            if not self.finalized:
                self._call_finalizers(response)

    def respond(
        self,
        status_code: int = 200,
        payload: t.Any = None,
        headers: t.Union[Headers, t.Mapping] = None,
    ):
        """
        Convenience method for handlers to stop the chain and set the given status and payload to the
        current response object.

        :param status_code: the HTTP status code
        :param payload: the payload of the response
        :param headers: additional headers
        """
        self.response.status_code = status_code
        if isinstance(payload, (list, dict)):
            self.response.set_json(payload)
        elif isinstance(payload, (str, bytes, bytearray)):
            self.response.data = payload
        elif payload is None and not self.response.response:
            self.response.response = []
        else:
            self.response.response = payload

        if headers:
            self.response.headers.update(headers)

        self.stop()

    def stop(self) -> None:
        """
        Stop the processing of the request handlers and proceed with response handlers.
        """
        self.stopped = True

    def terminate(self) -> None:
        """
        Terminate the handler chain, which skips response handlers.
        """
        self.terminated = True

    def throw(self, error: Exception) -> None:
        """
        Raises the given exception after the current request handler is done. This has no effect in response handlers.
        :param error: the exception to raise
        """
        self.error = error

    def _call_response_handlers(self, response):
        for handler in self.response_handlers:
            if self.terminated:
                return

            try:
                handler(self, self.context, response)
            except Exception as e:
                msg = "exception while running response handler"
                if LOG.isEnabledFor(logging.DEBUG):
                    LOG.exception(msg)
                else:
                    LOG.warning(msg + ": %s", e)

    def _call_finalizers(self, response):
        for handler in self.finalizers:
            try:
                handler(self, self.context, response)
            except Exception as e:
                msg = "exception while running request finalizer"
                if LOG.isEnabledFor(logging.DEBUG):
                    LOG.exception(msg)
                else:
                    LOG.warning(msg + ": %s", e)

    def _call_exception_handlers(self, e, response):
        for exception_handler in self.exception_handlers:
            try:
                exception_handler(self, e, self.context, response)
            except Exception as nested:
                # make sure we run all exception handlers
                msg = "exception while running exception handler"
                if LOG.isEnabledFor(logging.DEBUG):
                    LOG.exception(msg)
                else:
                    LOG.warning(msg + ": %s", nested)


class CompositeHandler:
    """
    A handler that sequentially invokes a list of Handlers, forming a stripped-down version of a handler
    chain.
    """

    handlers: list[Handler]

    def __init__(self, return_on_stop=True) -> None:
        """
        Creates a new composite handler with an empty handler list.

        TODO: build a proper chain nesting mechanism.

        :param return_on_stop: whether to respect chain.stopped
        """
        super().__init__()
        self.handlers = []
        self.return_on_stop = return_on_stop

    def append(self, handler: Handler) -> None:
        """
        Adds the given handler to the list of handlers.

        :param handler: the handler to add
        """
        self.handlers.append(handler)

    def remove(self, handler: Handler) -> None:
        """
        Remove the given handler from the list of handlers
        :param handler: the handler to remove
        """
        self.handlers.remove(handler)

    def __call__(self, chain: HandlerChain, context: RequestContext, response: Response):
        for handler in self.handlers:
            handler(chain, context, response)

            if chain.terminated:
                return
            if chain.stopped and self.return_on_stop:
                return


class CompositeExceptionHandler:
    """
    A exception handler that sequentially invokes a list of ExceptionHandler instances, forming a
    stripped-down version of a handler chain for exception handlers.
    """

    handlers: t.List[ExceptionHandler]

    def __init__(self) -> None:
        """
        Creates a new composite exception handler with an empty handler list.
        """
        self.handlers = []

    def append(self, handler: ExceptionHandler) -> None:
        """
        Adds the given handler to the list of handlers.

        :param handler: the handler to add
        """
        self.handlers.append(handler)

    def remove(self, handler: ExceptionHandler) -> None:
        """
        Remove the given handler from the list of handlers
        :param handler: the handler to remove
        """
        self.handlers.remove(handler)

    def __call__(
        self, chain: HandlerChain, exception: Exception, context: RequestContext, response: Response
    ):
        for handler in self.handlers:
            call_safe(
                handler,
                args=(chain, exception, context, response),
                exception_message="exception while running exception handler",
            )


class CompositeResponseHandler(CompositeHandler):
    """
    A CompositeHandler that by default does not return on stop, meaning that all handlers in the composite
    will be executed, even if one of the handlers has called ``chain.stop()``. This mimics how response
    handlers are executed in the ``HandlerChain``.
    """

    def __init__(self) -> None:
        super().__init__(return_on_stop=False)


class CompositeFinalizer(CompositeResponseHandler):
    """
    A CompositeHandler that uses ``call_safe`` to invoke handlers, so every handler is always executed.
    """

    def __call__(self, chain: HandlerChain, context: RequestContext, response: Response):
        for handler in self.handlers:
            call_safe(
                handler,
                args=(chain, context, response),
                exception_message="Error while running request finalizer",
            )



"""
DEPRECATED: use ``from rolo.routing import handler_dispatcher`` instead
"""
from .routing.handler import ResultValue, handler_dispatcher

__all__ = [
    "handler_dispatcher",
    "ResultValue",
]
